%%%----------------------------------------------------------------------
%%% File    : cd_storage.erl
%%% Author  : Sanchez Penas Juan Jose <juanjo@lfcia.org>
%%% Purpose : 
%%% Created : 3 Jun 2002 by Sanchez Penas Juan Jose
%%%----------------------------------------------------------------------

-module(cd_storage).
-author('juanjo@lfcia.org').

%%-compile(export_all).
%%-export([Function/Arity, ...]).

-behaviour(gen_server).

%% External exports
-export([start_link/5]).

%% gen_server callbacks
-export([init/1, handle_call/3]).

%%%----------------------------------------------------------------------
%%% API
%%%----------------------------------------------------------------------
start_link(Name, Movies, MaxConnections, MaxBandwidth, Cost) ->
    gen_server:start_link({local,Name}, ?MODULE, 
			  [Movies, MaxConnections, MaxBandwidth, Cost], []).

%%%----------------------------------------------------------------------
%%% Callback functions from gen_server
%%%----------------------------------------------------------------------

%%----------------------------------------------------------------------
%% Func: init/1
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          ignore               |
%%          {stop, Reason}
%%----------------------------------------------------------------------
init([Movies, MaxConnections, MaxBandwith, Cost]) ->
    io:format("CD Storage init~n"),
    {ok, {Movies,MaxConnections,MaxBandwith,Cost,0,0}}.

%%----------------------------------------------------------------------
%% Func: handle_call/3
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_call({lookup,MO,Profile}, From, 
	    {Movies, MaxConnections, MaxBandwidth, Cost, Connections, UsedBandwidth}) -> 
    {reply, 
     {lookupAns, 
      scheduling(MO, Profile, Movies,
		 UsedBandwidth, Connections, MaxConnections, MaxBandwidth, 
		 Cost, self())},
     {Movies, MaxConnections, MaxBandwidth, Cost, Connections, UsedBandwidth}};

handle_call({play, MO, Bandwidth}, From, 
	    {Movies, MaxConnections, MaxBandwidth, Cost, Connections, UsedBandwidth}) ->
    {reply, self(), {Movies, MaxConnections, MaxBandwidth, Cost, 
		     Connections + 1, UsedBandwidth + Bandwidth}}.

%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------

% This function returns a list in the form [{Bandwidth, CostForThisBandwidth, ProcId}]
% where each tuple represents an option: the MO requested by the user can be served from
% this device for the bw Bandwidth at a local cost CostForThisBandwidth, where the cost
% tries to represent 'how good or bad' is for the system to choose this option as provider

scheduling(MO, Profile, Movies, UsedBandwidth, UsedConnections, 
	   MaxConnections, MaxBandwidth, Cost, ProcId) -> 
   case UsedConnections >= MaxConnections of
     true -> 
         []; % All the available connections (channels) are already in use
     false ->
	   case UsedBandwidth >= MaxBandwidth of 
	       true ->
		   []; % All the available Bw is already in use
	       false ->
		   AvailableBw = check_available_bw(MO, Profile, Movies),
		   scheduling_still_resources(AvailableBw, UsedBandwidth, UsedConnections, 
					      MaxConnections, MaxBandwidth, Cost, ProcId)
	   end
   end.

% Once we know that the device has still some resources, we calculate the set
% of possible bandwidths that can be served

scheduling_still_resources ([], UsedBandwidth, UsedConnections, 
			    MaxConnections, MaxBandwidth, Cost, ProcId) -> 
    [];
scheduling_still_resources ([B|RestBandwidths], UsedBandwidth, UsedConnections, 
			    MaxConnections, MaxBandwidth, Cost, ProcId) ->
    case MaxBandwidth >= (UsedBandwidth+B) of
	false -> 
	    scheduling_still_resources (RestBandwidths, UsedBandwidth, UsedConnections, 
					MaxConnections, MaxBandwidth, Cost, ProcId);
	true -> [
		  {B,
		   cost(B,UsedBandwidth, UsedConnections, MaxConnections, MaxBandwidth, Cost),
		   ProcId}
		  |
		  scheduling_still_resources (RestBandwidths, UsedBandwidth, UsedConnections, 
					      MaxConnections, MaxBandwidth, Cost, ProcId)
		 ]
    end.
    
% Calculates the cost of serving the MO with the given bandwidth from thist device
%
% NOW THE COST IS CALCULATED AS PROPORTIONAL TO THE BANDWIDTH OF THE 

cost(MOBandwidth, UsedBandwidth, UsedConnections, MaxConnections, MaxBandwidth, Cost) ->
    MOBandwidth * (UsedConnections+1) * Cost.
    
% Movies (third argument) is a list of tuples of the form {MO, Profile}, 
% where Profile is a list of Bws (qualities) in which the movie is 
% available in that device
%
% Profile is a list of Bws (qualities) requested by the next level
% MO is the identifier of the movie requested by the next level

check_available_bw (MO, Profile, []) ->
    [];
check_available_bw (MO1, Profile, [{MO2,AvailableProfile}|OtherMOs]) ->
    case (MO1 == MO2) of
	true ->
	    list_intersection (Profile, AvailableProfile);    
	false ->
	    check_available_bw (MO1, Profile, OtherMOs)
    end.
    
% We are asuming that ordered lists are received !!

list_intersection ([],List) ->
    [];
list_intersection (List, []) ->
    [];
list_intersection ([Hd1|Tail1], [Hd2|Tail2]) -> % The same element
    case Hd1 >= Hd2 of
        false ->
            list_intersection (Tail1, [Hd2|Tail2]);
        true ->
            case Hd1 == Hd2 of
                 true ->
                   [Hd1 | list_intersection (Tail1, Tail2)];
                 false ->
                   list_intersection ([Hd1|Tail1], Tail2)
            end
    end.



